from binascii import hexlify, unhexlify
import unittest

from .makwa import Makwa, hashpw, checkpw


class MakwaTest(unittest.TestCase):

    def test_spec_vector(self):
        pi = unhexlify(
            '4765676F206265736877616A692761616B656E20617765206D616B77613B206F6'
            'E7A61616D206E616E69697A61616E697A692E'
        )
        n = int(
            'C22C40BBD056BB213AAD7C830519101AB926AE18E3E9FC9699C806E0AE5C2594'
            '14A01AC1D52E873EC08046A68E344C8D74A508952842EF0F03F71A6EDC077FAA'
            '14899A79F83C3AE136F774FA6EB88F1D1AEA5EA02FC0CCAF96E2CE86F3490F49'
            '93B4B566C0079641472DEFC14BECCF48984A7946F1441EA144EA4C802A457550'
            'BA3DF0F14C090A75FE9E6A77CF0BE98B71D56251A86943E719D27865A489566C'
            '1DC57FCDEFACA6AB043F8E13F6C0BE7B39C92DA86E1D87477A189E73CE8E311D'
            '3D51361F8B00249FB3D8435607B14A1E70170F9AF36784110A3F2E67428FC18F'
            'B013B30FE6782AECB4428D7C8E354A0FBD061B01917C727ABEE0FE3FD3CEF761',
            16
        )
        sigma = unhexlify('C72703C22A96D9992F3DEA876497E392')

        makwa = Makwa(work_factor=4096, pre_hashing=False)
        digest = makwa._digest(pi, n, salt=sigma)
        self.assertEqual(hexlify(digest), b'c9cea0e6ef09393ab1710a08')

        h = makwa.hash(pi, n, salt=sigma)
        self.assertEqual(h, '+RK3n5jz7gs_s211_xycDwiqW2ZkvPeqHZJfjkg_yc6g5u8JOTqxcQoI')

        h = hashpw(pi, n, salt=sigma, work_factor=4096, pre_hashing=False)
        self.assertEqual(h, '+RK3n5jz7gs_s211_xycDwiqW2ZkvPeqHZJfjkg_yc6g5u8JOTqxcQoI')
        self.assertEqual(h, '+RK3n5jz7gs_s211_xycDwiqW2ZkvPeqHZJfjkg_yc6g5u8JOTqxcQoI')

        self.assertTrue(makwa.check(pi, h, n))
        self.assertTrue(checkpw(pi, h, n))

        self.assertFalse(makwa.check(b'password', h, n))
        self.assertFalse(checkpw(b'password', h, n))

        self.assertFalse(makwa.check(pi, h, 0xbadc0de))
        self.assertFalse(checkpw(pi, h, 0xbadc0de))


if __name__ == '__main__':
    unittest.main()



from binascii import a2b_base64, b2a_base64, hexlify
from hashlib import sha256
import hmac
from os import urandom
from struct import pack


def int_to_bytes(i):
    bs = b''
    while i != 0:
        bs = pack('=B', i & 0xff) + bs
        i >>= 8
    return bs


def hashpw(password, n, salt=None, h=sha256, work_factor=4096, pre_hashing=True):
    makwa = Makwa(h=h, work_factor=work_factor, pre_hashing=pre_hashing)
    return makwa.hash(password, n, salt=salt)


def checkpw(password, hashed_password, n, h=sha256):
    state = hashed_password.split('_')[1]
    work_factor = int(state[1]) * (1 << int(state[2:]))
    pre_hashing = state[0] in 'rb'
    makwa = Makwa(h=h, work_factor=work_factor, pre_hashing=pre_hashing)
    return makwa.check(password, hashed_password, n)


class Makwa:
    def __init__(self, h=sha256, work_factor=4096, pre_hashing=True):
        self.h = h
        if work_factor == 0:
            raise ValueError('Work factor cannot be 0')
        self.m_cost = work_factor
        self.pre_hashing = pre_hashing
        self.post_hashing_length = 12

    def hash(self, password, n, salt=None,):
        if not isinstance(password, bytes):
            raise TypeError('Unicode-objects must be encoded before hashing')
        if salt is None:
            salt = urandom(16)

        h = ''
        h += self._base64(self._kdf(int_to_bytes(n), 8))
        h += '_'
        h += self._state_data()
        h += '_'
        h += self._base64(salt)
        h += '_'
        h += self._base64(self._digest(password, n, salt=salt))
        return h

    def check(self, password, hashed_password, n):
        if not isinstance(password, bytes):
            raise TypeError('Unicode-objects must be encoded before hashing')

        modhash, state, salt, digest = hashed_password.split('_')
        modhash = self._unbase64(modhash)
        salt = self._unbase64(salt)
        digest = self._unbase64(digest)

        if self._kdf(int_to_bytes(n), 8) != modhash:
            return False

        check_digest = self._digest(password, n, salt=salt)
        return digest == check_digest

    def _digest(self, password, n, salt=None,):
        if not isinstance(password, bytes):
            raise TypeError('Unicode-objects must be encoded before hashing')
        if salt is None:
            salt = urandom(16)

        if self.m_cost == 0:
            raise ValueError('Work factor cannot be 0')
        k = (n.bit_length() + 7) // 8
        if k < 160:
            raise ValueError('Modulus must be >= 160 bits')

        if self.pre_hashing:
            password = self._kdf(password, 64)

        u = len(password)
        if u > 255 or u > (k - 32):
            raise ValueError('Password is to long to be hashed under these parameters')
        sb = self._kdf(salt + password + pack('=B', u), k - 2 - u)
        xb = b'\x00' + sb + password + pack('=B', u)

        x = int(hexlify(xb), 16)
        for _ in range(self.m_cost + 1):
            x = pow(x, 2, n)
        out = int_to_bytes(x)

        if self.post_hashing_length not in [0, None]:
            out = self._kdf(out, self.post_hashing_length)

        return out

    def _kdf(self, data, out_len):
        r = self.h().digest_size
        V = b'\x01' * r
        K = b'\x00' * r
        K = hmac.new(K, msg=(V + b'\x00' + data), digestmod=self.h).digest()
        V = hmac.new(K, msg=V, digestmod=self.h).digest()
        K = hmac.new(K, msg=(V + b'\x01' + data), digestmod=self.h).digest()
        V = hmac.new(K, msg=V, digestmod=self.h).digest()
        T = b''
        while len(T) < out_len:
            V = hmac.new(K, msg=V, digestmod=self.h).digest()
            T += V
        return T[:out_len]

    def _base64(self, m):
        b64 = b2a_base64(m)
        if not isinstance(b64, str):
            b64 = b64.decode()
        return b64.replace('=', '').replace('\n', '')

    def _unbase64(self, m):
        padding = ((3 - (len(m) % 3)) % 3) * '='
        return a2b_base64(m + padding)

    def _state_data(self):
        ret = ''

        pre, post = self.pre_hashing, self.post_hashing_length
        if not pre and not post:
            ret += 's'
        elif pre and not post:
            ret += 'r'
        elif not pre and post:
            ret += 's'
        else:
            ret += 'b'

        delta, w = 0, self.m_cost
        while w & 1 == 0:
            delta += 1
            w //= 2

        ret += '2' if w == 1 else '3'
        ret += str(delta - 1) if w == 1 else str(delta)
        return ret



from setuptools import setup, Extension, Command


class TestCommand(Command):
    user_options = []
    description = "Run all tests"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from subprocess import call
        call(['python', '-m', 'makwa.test'])


setup(
    name='makwa',
    version='1.0.1',
    author='Anton Kueltz',
    author_email='kueltz.anton@gmail.com',
    keywords='makwa password hashing kdf',
    description='A password hashing function that supports delegation',
    long_description=''.join(open('README.rst', 'r').readlines()),
    url='https://github.com/AntonKueltz/makwa',
    packages=['makwa'],
    cmdclass={'test': TestCommand},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Security :: Cryptography',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)
